//
// LocoTool - A program to edit Locomotion object description .dat files
//
// Copyright (C) 2004 by Josef Drexler
//
// This file is distributed under the term of the GNU General Public Licence
// version 2 or higher.  Please read the file LICENCE for details.
//
//
// png.c - functions to read and write png files using libpng
//

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <png.h>

#include "common.h"
#include "dump.h"

int bgcol = 255;
png_byte trnscol = 0;

#ifdef DEBUGDMP
int colhist[256] = { -1 };
#endif

//
// Take a row of sprite data and assemble it into a PNG pixel row
//
// For sprites that are chunked, the "holes" in the data are transparent,
// and we use colour bgcol to indicate that.
//
// Sprites that aren't chunked are simply copied over.
//
uint32_t putspriterow(uint8_t *data, int flags, int y, int width, uint8_t *row)
{
  uint32_t offset, last, ofs, len;

#ifdef DEBUGDMP
  uint32_t x1;

  if (colhist[0] == -1)
	memset(colhist, 0, sizeof(colhist));
#endif

  memset(row, bgcol, width);

  if (!(flags&4)) {
	memcpy(row, data+y*width, width);

#ifdef DEBUGDMP
	for (x1=0; x1<width; x1++)
		colhist[row[x1]]++;
#endif

	return (y+1)*width;
  }

  offset=getvalue(data+2*y, 2);

  do {
	last=data[offset]  & 0x80;
	len=data[offset++] & 0x7f;
	ofs=data[offset++];

	memcpy(row+ofs, data+offset, len);

#ifdef DEBUGDMP
	for (x1=ofs; x1<ofs+len; x1++)
		colhist[row[x1]]++;
#endif

	offset+=len;
  } while (!last);

  return offset;
}


//
// Reverse of the above, take a row of PNG pixels and encode it as sprite data
//
// For chunked sprites, we also assemble the data offset array
//
uint32_t getspriterow(uint8_t *data, uint32_t size, int flags, int y, int width, uint8_t *row)
{
  uint32_t offset, x1, x2, len;

  if (!(flags&4)) {
	memcpy(data+y*width, row, width);
	return (y+1)*width;
  }

  ((uint16_t*) data)[y] = size;

  x1=x2=0;

  offset = size;
  while (x1 < width) {
	while ( (x1 < width) && (row[x1]==bgcol))
		x1++;

	if (x1 < width) {
		len = 1;
		x2 = x1+len;

		while ((x2 < width) && (len < 127) && (row[x2]!=bgcol)) {
			x2++;
			len++;
		}

		offset = size;
		data[size++] = len;
		data[size++] = x1;

		memcpy(data+size, row+x1, len);
		size += len;

	} else {	// transparent to end of line
		if (x2 == 0) {	// totally empty line
			data[size++] = 0;
			data[size++] = 0;
		}
		x2 = x1;
	}
	x1 = x2;
  }

  data[offset] |= 0x80;

  return size;
}

//
// IO functions called by the PNG writer
//
void xml_write(png_structp png_ptr, png_bytep data, png_size_t size)
{
  fwrite(data, 1, size, png_ptr->io_ptr);
}

void xml_flush(png_structp png_ptr)
{
  fflush(png_ptr->io_ptr);
}

#include "locopal.h"

//
// Write a sprite to the given PNG file
//
uint32_t makepng(FILE *xml, FILE *png, uint8_t *data, uint32_t width, uint32_t height, uint32_t flags)
{
   png_structp png_ptr;
   png_infop info_ptr;
   png_colorp palette;
   png_byte *row;
   png_color_16 backgnd;

   int i, j, ccolind;
   uint32_t dumped;

   dumped = 0;

#ifdef __windows
	// make sure the CR<->CRLF translation is off
   if (setmode(fileno(xml), O_BINARY) == 1) die("Can't switch to binary file mode");
#endif

	// create the png structures
   png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
   if (!png_ptr) die("Error creating png write struct");

   info_ptr = png_create_info_struct(png_ptr);
   if (!info_ptr) {
	png_destroy_write_struct(&png_ptr, png_infopp_NULL);
	die("Error creating png info struct");
   }

	// error handlingg
   if (setjmp(png_jmpbuf(png_ptr))) {
		// if we get here, we had a problem writing the file
	png_destroy_write_struct(&png_ptr, &info_ptr);
	die("Error while writing png data");
   }

   if (embedpng) {
		// use our own write functions to write the data to the
		// xml file instead
	png_set_write_fn(png_ptr, xml, xml_write, xml_flush);
   } else if (allinone) {
   } else {
	png_init_io(png_ptr, png);
   }

	// set image dimensions
   png_set_IHDR(png_ptr, info_ptr, width, height, 8, PNG_COLOR_TYPE_PALETTE,
      PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

	// set the palette
   palette = (png_colorp)png_malloc(png_ptr, PNG_MAX_PALETTE_LENGTH * sizeof (png_color));
   if (!palette) die("Not enough memory for %s, wanted %d", "palette", PNG_MAX_PALETTE_LENGTH * sizeof (png_color));

   for (i=0; i<256; i++) {
	palette[i].red = locopalette[i][0];
	palette[i].green = locopalette[i][1];
	palette[i].blue = locopalette[i][2];
   }

	// company colours, set which colour to use
   ccolind = (palettetype&~1)*6-2;
   if (ccolind > 0) {
	for (i=0; i<12; i++) {
		palette[companycolours[i]].red = locopalette[ccolind+i][0];
		palette[companycolours[i]].green = locopalette[ccolind+i][1];
		palette[companycolours[i]].blue = locopalette[ccolind+i][2];
	}

		// disable this colour in non-company-colour range?
	if (palettetype & 1) {
		for (i=0; i<12; i++) {
			palette[ccolind+i].red = locopalette[companycolours[i]][0];
			palette[ccolind+i].green = locopalette[companycolours[i]][1];
			palette[ccolind+i].blue = locopalette[companycolours[i]][2];
		}
	}
   }

   png_set_PLTE(png_ptr, info_ptr, palette, PNG_MAX_PALETTE_LENGTH);

	// set the gamma
   png_set_gAMA(png_ptr, info_ptr, 1.4);

	// set the transparent colour
   png_set_tRNS(png_ptr, info_ptr, &trnscol, 1, NULL);

	// and the background colour
   backgnd.red = backgnd.green = backgnd.blue = 65535;
   png_set_bKGD(png_ptr, info_ptr, &backgnd);

	// write the png header
   png_write_info(png_ptr, info_ptr);

   if (embedpng || !allinone) {

	   row = my_alloc(width, "row");

		// write the image line by line
	   for (i=0; i<height; i++) {
		uint32_t offset = putspriterow(data, flags, i, width, row);

		if (offset > dumped) dumped=offset;

		png_write_row(png_ptr, row);
	   }

	   my_free(row);

   }

	// done writing, clean up
   png_write_end(png_ptr, info_ptr);
   png_free(png_ptr, palette);
   png_destroy_write_struct(&png_ptr, &info_ptr);

#ifdef __windows
	// switch back to text mode for the remainder of the XML
   if (setmode(fileno(xml), O_TEXT) == 1) die("Can't switch to text file mode");
#endif

   return dumped;
}

//
// Read a png file, return its data and properties
//
uint32_t readpng(const char *name, uint8_t **data, int flags, uint32_t *width, uint32_t *height)
{
   png_structp png_ptr;
   png_infop info_ptr;
   int bit_depth, color_type, interlace_type;
   FILE *png;
   int i;
   uint8_t *sprite;
   uint32_t maxsize, size;
   png_byte **rows;
   png_colorp palette;

   png = fopen(name, "rb");
   if (!png) die("Can't read %s: %s", name, strerror(errno));

   png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
   if (!png_ptr) {
	fclose(png);
	die("Error creating png read struct");
   }

   info_ptr = png_create_info_struct(png_ptr);
   if (!info_ptr) {
	fclose(png);
	png_destroy_read_struct(&png_ptr, png_infopp_NULL, png_infopp_NULL);
	die("Error creating png info struct");
   }

   if (setjmp(png_jmpbuf(png_ptr)))
   {
      png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
      fclose(png);
	die("Error while reading png data");
   }

   png_init_io(png_ptr, png);

   png_read_png(png_ptr, info_ptr, 0, png_voidp_NULL);

   *height = png_get_image_height(png_ptr, info_ptr);
   *width = png_get_image_width(png_ptr, info_ptr);

   if (png_get_bit_depth(png_ptr, info_ptr) != 8)
	die("Unsupported PNG bit depth in %s", name);

   maxsize = (*height) * (*width)*3 + 2 * (*height) + 32;
   sprite = my_alloc(maxsize, "sprite");

   bgcol = 255;
   if (png_get_color_type(png_ptr, info_ptr) & PNG_COLOR_MASK_PALETTE) {
	png_get_PLTE(png_ptr, info_ptr, &palette, &i);
	if ( (palette[255].red != 255) || (palette[255].green != 255) || (palette[255].blue != 255) )
		bgcol = 0;
   }

   rows = png_get_rows(png_ptr, info_ptr);
   size = 0;
   if (flags & 4)
	size = *height*2;

   for (i=0; i<*height; i++) {
	size = getspriterow(sprite, size, flags, i, *width, rows[i]);
	if (size >= maxsize) die("Sprite buffer overflow in %s", name);
   }

   sprite = my_realloc(sprite, size, maxsize, "sprite");
   *data = sprite;

   // clean up
   png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);

   fclose(png);

   return size;
}
